/*
 * Created on May 18, 2005
 *
 */
package com.sunlabs.archsim.build;

import java.lang.reflect.Constructor;
import java.util.Iterator;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.classRegistry.ClassRegistry;
import com.sunlabs.archsim.engine.Director;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.gasp.Component;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.globals.Globals;

/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author ib27688
 * 
 * Build a GASP model from a model description.
 */
public class GASPModelBuilder {
	
	protected Globals globals= Globals.getInstance();

	ModelDescription modDes;
	
	public GASPModelBuilder() {
		modDes= null;
	} // ModelBuilder constructor

	/** 
	 * Build components and connections from the model description
	 */
	public CompositeEntity buildModel(ModelDescription m) throws ModelNotBuiltException {
		// use model name for the "top" group
		modDes= m;
		CompositeEntity modelGroup= new CompositeEntity(modDes.name);
		
		// build components
		Iterator ci= modDes.componentNameMap.values().iterator();
		while (ci.hasNext()) {
			ComponentRecord cr= (ComponentRecord)ci.next();
			
			// get a constructor	
			try {
				
/*				if (cr.attributes == null) {

					// the constructor will have only one parameter
					Constructor ctor= cr.type.getConstructor(
							new Class[] {String.class});
					
					cr.component= (Component)ctor.newInstance(
							new Object[] { cr.name });
					
				}
				else {
*/
				// require existence of a two parameter constructor
				Constructor ctor= cr.type.getDeclaredConstructor(
						new Class[] {String.class, Parameters.class});
				
				// System.out.println(cr.type.getName() + ": constructor found");
				// System.out.println(cr.attributes);
				
				cr.component= (Component)ctor.newInstance(
						new Object[] {cr.name, cr.attributes});					

				// System.out.println(cr.type.getName() + ": object made");

				
				modelGroup.addMember(cr.component);
			}
			catch (Exception e) {
				System.err.println(e.getClass().getName());
				e.printStackTrace();
				// System.out.println(e.getMessage());
				throw new ModelNotBuiltException("Error: Component "
						+ cr.name + " of type "
						+ cr.type.getName() 
						+ " cannot be built with a constructor that"
						+ " takes String and Parameters as arguments");
			} // catch
		} // while ci has next
		
		// now build connections
		
		// start by figuring out what to log
		Boolean lds= globals.booleanValue(BuildDefaults.LOG_CONNNECTION_DATA_STATE);
		boolean logDataState= (lds != null)
			? lds
			: BuildDefaults.LOG_CONNNECTION_DATA_STATE_DEF;
		
		Boolean las= globals.booleanValue(BuildDefaults.LOG_CONNNECTION_ACK_STATE);
		boolean logAckState= (las != null)
		? las
		: BuildDefaults.LOG_CONNNECTION_ACK_STATE_DEF;
		
		
		ci= modDes.connectionRecordList.iterator();
		while (ci.hasNext()) {
			// get a connection
			ConnectionRecord cr= (ConnectionRecord) ci.next();
			
//			System.out.println("\n\n\n" + cr + "\n\n\n");
			
			// make two states - one for data one for acknowledgments
			State dataState= new State(cr.name, modelGroup);
			State ackState= new State(BuildDefaults.ACK+cr.name, modelGroup);
			// both states journal their activity under the name
			// of the connection
			dataState.setJournalName(cr.name);
			ackState.setJournalName(cr.name);

			
			// which states must be initially active?
			if (cr.dataActive) {
				
				// data state is initially active
				// what is the value
				Object value= null;
				if (cr.dataValue != null 
						&& !cr.dataValue.equals("") 
						&& !cr.dataValue.equals("null")) {
					
					boolean typeFound= false;
					// is it integer?
					try {
						value= Integer.parseInt(cr.dataValue);
						typeFound= true;
					}
					catch (Exception e) {
						// empty catch
					}
					
					// is it boolean?
					if (!typeFound) {
						try {
							value= Boolean.parseBoolean(cr.dataValue);
						}
						catch (Exception e) {
							// empty catch
						}
					}

					// not an integer and boolean, assume string
					if (!typeFound) {
						value= new String(cr.dataValue);
					}			
				} // if not null
				
				// we have the value, set the data state initially active
				dataState.setInitiallyActive(new Data(value));
//				System.out.println("\n\n\n VALUE= " + value + "\n\n\n\n");

			}
			else {
				// ack state is initially active
				ackState.setInitiallyActive(null);				
			}
			
			
			
			// connect sources first
			Iterator si= cr.sources.iterator();
			while (si.hasNext()) {
				// get a source connection point
				ConnectionPoint cp= (ConnectionPoint)si.next();
				// get the source component
				// ASSUME the parser has ensured that the components exist
				ComponentRecord cmpRec= (ComponentRecord)modDes.componentNameMap.get(cp.componentName);
				Component cmp= cmpRec.component;
								
				// get the output terminal - the source of data
				Terminal out= cmp.getTerminal(cp.terminalName);
				// is the terminal OK?
				if (out == null || !(out instanceof OutputTerminal)) {
					// terminal is not good
					throw new ModelNotBuiltException("Error: Output terminal "
							+ cp.terminalName 
							+ " does not exist in component "  
							+ cmp.getName() 
							+ ", of class " 
							+ cmp.getClass().getName());
				}
				
				// now get the input terminal - for the acknowldegment
				Terminal ackOut= cmp.getTerminal(BuildDefaults.ACK+cp.terminalName);
				// is the terminal OK?
				if (ackOut == null || !(ackOut instanceof InputTerminal)) {
					// terminal is not good
					throw new ModelNotBuiltException("Error: Input terminal "
							+ BuildDefaults.ACK+cp.terminalName 
							+ " does not exist in component "  
							+ cmp.getName() 
							+ ", of class " 
							+ cmp.getClass().getName());
				}
				
				// connect the state and the terminals
				dataState.attach(out);
				ackState.attach(ackOut);
				
			} // while there are sources
			
			// connect destinations 
			Iterator di= cr.destinations.iterator();
			while (di.hasNext()) {
				// get a destination connection point
				ConnectionPoint cp= (ConnectionPoint)di.next();
				// get the destination component
				// ASSUME the parser has ensured that the components exist
				ComponentRecord cmpRec= (ComponentRecord)modDes.componentNameMap.get(cp.componentName);
				Component cmp= cmpRec.component;
				
				// get the output terminal - the source of data
				Terminal in= cmp.getTerminal(cp.terminalName);
				// is the terminal OK?
				if (in == null || !(in instanceof InputTerminal)) {
					// terminal is not good
					throw new ModelNotBuiltException("Error: Input terminal "
							+ cp.terminalName 
							+ " does not exist in component "  
							+ cmp.getName() 
							+ ", of class " 
							+ cmp.getClass().getName());
				}
				
				// now get the input terminal - for the acknowldegment
				Terminal ackIn= cmp.getTerminal(BuildDefaults.ACK+cp.terminalName);
				// is the terminal OK?
				if (ackIn == null || !(ackIn instanceof OutputTerminal)) {
					// terminal is not good
					throw new ModelNotBuiltException("Error: Output terminal "
							+ BuildDefaults.ACK+cp.terminalName 
							+ " does not exist in component "  
							+ cmp.getName() 
							+ ", of class " 
							+ cmp.getClass().getName());
				}
				
				// connect the state and the terminals
				dataState.attach(in);
				ackState.attach(ackIn);
				
			} // while there are sources
			
			
			// set the journaling
			dataState.setJournal(logDataState);
			ackState.setJournal(logAckState);
			
			// set the ack-state active
			// was set active at creation
			// ackState.setActive(null);
			
			
			
		} // while there are connections
		
		return modelGroup;
		
	} // build model

	
	public static void main(String[] args) {
		
		Director director= Director.getInstance();
		
		ClassRegistry reg=  ClassRegistry.instance;
		reg.load("../classRegistry/testRegLoader.xml");
		
		
		ModelDescription model= new ModelDescription();
		model.load("testModelLoader.xml");
		
		model.print();
		
		director.info("\n\n");
		
		GASPModelBuilder builder= new GASPModelBuilder();
		
		CompositeEntity grp;
		
		try {
			grp= builder.buildModel(model);
			
			grp.print();
		}
		catch (ModelNotBuiltException e) {
			director.error(e.getMessage());
			e.printStackTrace();
		}
    			
	} // main
	
} // ModelBulder
